/*-----------------------------------------------------------------------*/
/* MMC/SDC (in SPI mode) control module  (C)ChaN, 2007                   */
/*-----------------------------------------------------------------------*/
/* Only rcvr_spi(), xmit_spi(), disk_timerproc() and some macros         */
/* are platform dependent.                                               */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Receive a byte from MMC via SPI  (Platform dependent)                 */
/*-----------------------------------------------------------------------*/

unsigned char rcvr_spi (void)
{ 

	S0SPCR = 0x0820; // 8 bitow

	S0SPDR=0xFF;
  while ((S0SPSR & 0x80)==0);
  return S0SPDR;
}

/*-----------------------------------------------------------------------*/
/* Wait for card ready                                                   */
/*-----------------------------------------------------------------------*/

unsigned char wait_ready (void)
{
	BYTE res;

	S0SPCR = 0x0820; // 8 bitow

	Timer2xx = 0x07FFFFFF; //50;	/* Wait for ready in timeout of 500ms */
	rcvr_spi();
	do
		res = rcvr_spi();
	while ((res != 0xFF) && Timer2xx--);

	return res;
}

/*-----------------------------------------------------------------------*/
/* Receive a data packet from MMC                                        */
/*-----------------------------------------------------------------------*/

unsigned char rcvr_datablock (
	BYTE *buff,			/* Data buffer to store received data */
	unsigned int btr			/* Byte count (must be even number) */
)
{
	unsigned char token;
	int f, xx;
  
	S0SPCR = 0x0820; // 8 bitow

	Timer1xx = 0x01FFFFFF; //10;
	do {							/* Wait for data packet in timeout of 100ms */
		token = rcvr_spi();
	} while ((token == 0xFF) && Timer1xx--);
	if(token != 0xFE) return FALSE;	/* If not valid data token, retutn with error */
	if(btr == 512) // -- ZP -- 
	{
		S0SPCR = 0x0024; // 16 bitow
		for(f=0;f<512;f=f+8)
		{
			S0SPDR = 0xFFFF; // Czytaj
			while(!(S0SPSR & 0x80)); // wait for Transit complete
			xx = S0SPDR; // Read dane 16b
			S0SPDR = 0xFFFF; //  Czytaj nastepne dane [A]
		*buff++ = xx >> 8; // Dane L
			*buff++ = xx; // H
	// - 2, 3 -----------------------------------------------
			// [A]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
			xx = S0SPDR; // Read dane 16b
			S0SPDR = 0xFFFF; //  Czytaj nastepne dane [B]
			*buff++ = xx >> 8; // Dane L
			*buff++ = xx; // H
	// - 4, 5 -----------------------------------------------
			// [B]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
			xx = S0SPDR; // Read dane 16b
			S0SPDR = 0xFFFF; //  Czytaj nastepne dane [C]
		*buff++ = xx >> 8; // Dane L
			*buff++ = xx; // H
	// - 6, 7 -----------------------------------------------
		// [ C]
		while(!(S0SPSR & 0x80)); // wait for Transit complete
			xx = S0SPDR; // Read dane 16b
		*buff++ = xx >> 8; // Dane L
			*buff++ = xx; // H
		}
		S0SPCR = 0x0820; // 8 bitow
	}
	else
		do {							/* Receive the data block into buffer */
		rcvr_spi_m(buff++);
		rcvr_spi_m(buff++);
	  } while (btr -= 2);
	rcvr_spi();						/* Discard CRC */
	rcvr_spi();

	return TRUE;					/* Return with success */
}

/*-----------------------------------------------------------------------*/
/* Send a data packet to MMC                                             */
/*-----------------------------------------------------------------------*/

#if _READONLY == 0
unsigned char xmit_datablock (
	const BYTE *buff,	/* 512 byte data block to be transmitted */
	unsigned char token			/* Data/Stop token */
)
{
	BYTE resp;
	unsigned long f, xx, yy; // -- ZP --
	uintptr_t *ptr_32;

	if (wait_ready() != 0xFF) return FALSE;

	S0SPCR = 0x0820; // 8 bitow

	xmit_spi(token);					/* Xmit data token */
	if (token != 0xFD) {	/* Is data token */
// oryginal
//		do {							/* Xmit the 512 byte data block to MMC */
//			xmit_spi(*buff++);
//			xmit_spi(*buff++);
//		} while (--wc);
// -- ZP ------------------------------------------------------
	  xx = (unsigned long)buff;
		ptr_32 = (uintptr_t*)xx; // uintptr_t
		S0SPCR = 0x0024; // 16 bitow
	  if(DISK_32_B != 0x5A63A653)
		 for(f=0;f<512;f=f+8)
		 {
			// Bufor *char
			xx = *buff++ << 8; // Dane L
			xx = xx + *buff++; // H
			S0SPDR = xx; // Zapisz
			xx = *buff++ << 8; // Dane L next [A]
			xx = xx + *buff++; // H next [A]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
	 // - 2, 3 -----------------------------------------------
			S0SPDR = xx; // Zapisz [A]
			xx = *buff++ << 8; // Dane L [B]
			xx = xx + *buff++; // H [B]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
	 // - 4, 5 -----------------------------------------------
			S0SPDR = xx; // Zapisz [B]
			xx = *buff++ << 8; // Dane L [C]
			xx = xx + *buff++; // H [C]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
	 // - 6, 7 -----------------------------------------------
			S0SPDR = xx; // Zapisz [C]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
		 }
		else
		 for(f=0;f<512;f=f+8)
		 {
			// Bufor *long
			yy = *ptr_32++;
			buff++;
			buff++;
			buff++;
			buff++;
			xx = yy & 0x0000FFFF; // Dane H, L
			xx = xx << 8; // Dane L
      xx = xx | (yy >> 8); // H
			S0SPDR = xx; // Zapisz
			xx = yy >> 24; // Dane H next [A]
			xx = xx | ((yy >> 16) & 0x0000FF00); // Dane L
			while(!(S0SPSR & 0x80)); // wait for Transit complete
	 // - 2, 3 -----------------------------------------------
			S0SPDR = xx; // Zapisz [A]
			yy = *ptr_32++;
      buff++;
      buff++;
      buff++;
      buff++;
			xx = yy & 0x0000FFFF; // Dane H, L [B]
			xx = xx << 8; // Dane L
			xx = xx | (yy >> 8); // H
			while(!(S0SPSR & 0x80)); // wait for Transit complete
	 // - 4, 5 -----------------------------------------------
			S0SPDR = xx; // Zapisz [B]
			xx = yy >> 24; // Dane H next [C]
      xx = xx | ((yy >> 16) & 0x0000FF00); // Dane L
			while(!(S0SPSR & 0x80)); // wait for Transit complete
	 // - 6, 7 -----------------------------------------------
			S0SPDR = xx; // Zapisz [C]
			while(!(S0SPSR & 0x80)); // wait for Transit complete
		 }
		S0SPCR = 0x0820; // 8 bitow

// ------------------------------------------------------------
		xmit_spi(0xFF);					/* CRC (Dummy) */
		xmit_spi(0xFF);
		resp = rcvr_spi();				/* Reveive data response */
		if ((resp & 0x1F) != 0x05)		/* If not accepted, return with error */
			return FALSE;
	}

	return TRUE;
}
#endif /* _READONLY */



/*-----------------------------------------------------------------------*/
/* Send a command packet to MMC                                          */
/*-----------------------------------------------------------------------*/

unsigned char send_cmd (
	BYTE cmd,		/* Command byte */
	DWORD arg		/* Argument */
)
{
	BYTE n, res;


	if (wait_ready() != 0xFF) return 0xFF;

	/* Send command packet */
	xmit_spi(cmd);						/* Command */
	xmit_spi((BYTE)(arg >> 24));		/* Argument[31..24] */
	xmit_spi((BYTE)(arg >> 16));		/* Argument[23..16] */
	xmit_spi((BYTE)(arg >> 8));			/* Argument[15..8] */
	xmit_spi((BYTE)arg);				/* Argument[7..0] */
	n = 0;
	if (cmd == CMD0) n = 0x95;			/* CRC for CMD0(0) */
	if (cmd == CMD8) n = 0x87;			/* CRC for CMD8(0x1AA) */
	xmit_spi(n);

	/* Receive command response */
	if (cmd == CMD12) rcvr_spi();		/* Skip a stuff byte when stop reading */
	n = 10;								/* Wait for a valid response in timeout of 10 attempts */
	do
		res = rcvr_spi();
	while ((res & 0x80) && --n);

	return res;			/* Return with the response value */
}


/*--------------------------------------------------------------------------

   Public Functions

---------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Initialize Disk Drive                                                 */
/*-----------------------------------------------------------------------*/

//DSTATUS disk_initialize (
unsigned char disk_initialize (
	BYTE drv		/* Physical drive nmuber (0) */
)
{
	BYTE n, f, ocr[4];
//    printf("Inside init: %x\n\r\n\r", Stat);
    Stat = Stat & ~STA_NODISK;
    
	if (drv) return STA_NOINIT;			/* Supports only single drive */
	if (Stat & STA_NODISK) return Stat;	/* No card in the socket */
    
//printf("after if\n\r");
	
	S0SPCCR = SPI_LOW; //SPI_LOW
//power_on();							/* Force socket power on */
	for (n = 10; n; n--) rcvr_spi();	/* 80 dummy clocks */

	SELECT();				/* CS = L */
	f = 0;
	if (send_cmd(CMD0, 0) == 1) {			/* Enter Idle state */
		Timer1xx = 0x3FFFFFFF; //100;						/* Initialization timeout of 1000 msec */
		if (send_cmd(CMD8, 0x1AA) == 1) {	/* SDC Ver2+ */
			for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();
			if (ocr[2] == 0x01 && ocr[3] == 0xAA) {	/* The card can work at vdd range of 2.7-3.6V */
				do {
					if (send_cmd(CMD55, 0) <= 1 && send_cmd(CMD41, 1UL << 30) == 0)	break;	/* ACMD41 with HCS bit */
				} while (Timer1xx--);
				if (Timer1xx && send_cmd(CMD58, 0) == 0) {	/* Check CCS bit */
					for (n = 0; n < 4; n++) ocr[n] = rcvr_spi();
					f = (ocr[0] & 0x40) ? 3 : 1;
				}
			}
		} else {							/* SDC Ver1 or MMC */
			if (send_cmd(CMD55, 0) <= 1 && send_cmd(CMD41, 0) <= 1) f = 4;	/* SDC? */
			do {
				if (!f && send_cmd(CMD1, 0) == 0) break;								/* CMD1 */
				if ( f && send_cmd(CMD55, 0) <= 1 && send_cmd(CMD41, 0) == 0) break;	/* ACMD41 */
			} while (Timer1xx--);
			if (Timer1xx) f = 1;
		}
		if (f == 1 && send_cmd(CMD16, 512) != 0)	/* Select R/W block length */
			f = 0;
	}
	DESELECT();			/* CS = H */
	rcvr_spi();			/* Idle (Release DO) */

	if (f & 1) {		/* Initialization succeded */
		BlkAddr = (f & 2) ? 1 : 0;	/* Get addressing mode */
		Stat &= ~STA_NOINIT;		/* Clear STA_NOINIT */
	} else {			/* Initialization failed */
		//power_off(); 
//		printf("powering off\n\r");
     return 0x7F; // failed
	}

    //set to fclk;
    S0SPCCR = cur_spi_spd; //SPI_HIGH
	return 0; // OK
}


/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

unsigned char disk_read (
	BYTE drv,			/* Physical drive nmuber (0) */
	char *buff,			/* Pointer to the data buffer to store read data */
	DWORD sector,		/* Start sector number (LBA) */
	BYTE count			/* Sector count (1..255) */
)
{
  counts = 0; // ZP
	if (drv || !count) return 0x7E;
	if (Stat & STA_NOINIT) return 0x7E;

	if (!BlkAddr) sector = sector << 9;	/* Convert to byte address if needed */

	SELECT();			/* CS = L */

	if (count == 1) {	/* Single block read */
		if ((send_cmd(CMD17, sector) == 0)	/* READ_SINGLE_BLOCK */
			&& rcvr_datablock(buff, 512))
			count = 0;
		else
			counts = 1; // ZP
	}
	else {				/* Multiple block read */
		if (send_cmd(CMD18, sector) == 0) {	/* READ_MULTIPLE_BLOCK */
			do {
				if (!rcvr_datablock(buff, 512)) break;
        counts++; // ZP
				buff += 512;
			} while (--count);
			send_cmd(CMD12, 0);				/* STOP_TRANSMISSION */
		}
	}

	DESELECT();			/* CS = H */
	rcvr_spi();			/* Idle (Release DO) */

	return count ? 0x7E : 0;
}



/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if _READONLY == 0
unsigned char disk_write (
	BYTE drv,			/* Physical drive nmuber (0) */
	char *buff,	/* Pointer to the data to be written */
	DWORD sector,		/* Start sector number (LBA) */
	BYTE count			/* Sector count (1..255) */
)
{
  counts = 0; // ZP
	if (drv || !count) return 0x7E;
	if (Stat & STA_NOINIT) return 0x7E;
	if (Stat & STA_PROTECT) return 0x7E;

	if (!BlkAddr) sector = sector << 9; // *= 512;	Convert to byte address if needed

	SELECT();			/* CS = L */

	if (count == 1) {	/* Single block write */
		if ((send_cmd(CMD24, sector) == 0)	/* WRITE_BLOCK */
			&& xmit_datablock(buff, 0xFE))
			count = 0;
		else
			counts = 1; // ZP
	}
	else {				/* Multiple block write */
		if (send_cmd(CMD25, sector) == 0) {	/* WRITE_MULTIPLE_BLOCK */
			do {
				if (!xmit_datablock(buff, 0xFC)) break;
        counts++; // ZP
				buff += 512;
			} while (--count);
			if (!xmit_datablock(0, 0xFD))	/* STOP_TRAN token */
				count = 1;
		}
	}

	DESELECT();			/* CS = H */
	rcvr_spi();			/* Idle (Release DO) */

	return count ? 0x7E : 0;
}
#endif /* _READONLY */



/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

unsigned char disk_ioctl (
	BYTE drv,		/* Physical drive nmuber (0) */
	BYTE ctrl,		/* Control code */
	void *buff		/* Buffer to send/receive control data */
)
{
	BYTE res;
	//BYTE n, csd[16], *ptr = buff;  
	BYTE n, csd[16]; 
	BYTE *ptr;
	int16_t csize;
	//int csize;
	ptr = buff;


	if (drv) return 0x7E;

	res = 0x7E;

	if (ctrl == CTRL_POWER) {
		switch (*ptr) {
		case 0:		/* Sub control code == 0 (POWER_OFF) */
			break;
		case 1:		/* Sub control code == 1 (POWER_ON) */
			break;
		case 2:		/* Sub control code == 2 (POWER_GET) */
			break;
		default :
			res = 0x7E;
		}
	}
	else {
		if (Stat & STA_NOINIT) return 0x7E;

		SELECT();		/* CS = L */

		switch (ctrl) {
		case GET_SECTOR_COUNT :	/* Get number of sectors on the disk (DWORD) */
			if ((send_cmd(CMD9, 0) == 0) && rcvr_datablock(csd, 16)) {
				n = csd[0] >> 6;
				SD_FILE_FORMAT = (csd[14]>>2) & 0x03; // CSD[11, 10]
				switch (n) {
				case 0 :	/* CSD ver 1.XX */
					n = (csd[5] & 15) + ((csd[10] & 128) >> 7) + ((csd[9] & 3) << 1) + 2;
					csize = (csd[8] >> 6) + ((int16_t)csd[7] << 2) + ((int16_t)(csd[6] & 3) << 10) + 1;
					*(DWORD*)buff = (DWORD)csize << (n - 9);
					res = RES_OK;
					break;
				case 1 :	/* CSD ver 2.00 */
					csize = csd[9] + ((int16_t)csd[8] << 8) + 1;
					*(DWORD*)buff = (DWORD)csize << 10;
					res = RES_OK;
					break;
				}
			}
			break;

		case GET_SECTOR_SIZE :	/* Get sectors on the disk (WORD) */
			*(WORD*)buff = 512;
			res = RES_OK;
			break;

		case CTRL_SYNC :	/* Make sure that data has been written */
			if (wait_ready() == 0xFF)
				res = RES_OK;
			break;

		case MMC_GET_CSD :	/* Receive CSD as a data block (16 bytes) */
			if (send_cmd(CMD9, 0) == 0		/* READ_CSD */
				&& rcvr_datablock(ptr, 16))
				res = RES_OK;
			break;

		case MMC_GET_CID :	/* Receive CID as a data block (16 bytes) */
			if (send_cmd(CMD10, 0) == 0		/* READ_CID */
				&& rcvr_datablock(ptr, 16))
				res = RES_OK;
			break;

		case MMC_GET_OCR :	/* Receive OCR as an R3 resp (4 bytes) */
			if (send_cmd(CMD58, 0) == 0) {	/* READ_OCR */
				for (n = 0; n < 4; n++)
					*ptr++ = rcvr_spi();
				res = RES_OK;
			}
			break;

		default:
			res = 0x7E;
		}

		DESELECT();			/* CS = H */
		rcvr_spi();			/* Idle (Release DO) */
	}

	return res;
}



/*-----------------------------------------------------------------------*/
